@@ -1,7 +1,9 @@
 from collections.abc import Generator, Iterable, Iterator
 from inspect import Parameter, cleandoc, signature
 from textwrap import dedent, fill, indent
-from typing import Mapping, Optional, Union
+from typing import Mapping, Optional, Sequence, Union
+
+from numpy.typing import ArrayLike, DTypeLike
 
 try:
     from typing import get_args as typing_get_args
@@ -19,13 +21,36 @@ class DocumentationError(Exception):
 
 
 def format_paragraph(s):
-    return fill(dedent(s.strip())) + newline
+    return fill(dedent(s.strip(newline))) + newline
 
 
 def format_indented_paragraph(s):
     return indent(format_paragraph(s), prefix="    ")
 
 
+def format_paragraphs(s):
+    prep = dedent(s.strip(newline))
+    paragraphs = prep.split(newline + newline)
+    docstring = ""
+    for paragraph in paragraphs:
+        if (
+            paragraph.startswith(" ")
+            or paragraph.startswith("..")
+            or paragraph.startswith(">")
+            or paragraph.startswith("[")
+        ):
+            # leave this as-is
+            docstring += paragraph + newline + newline
+        else:
+            # fill
+            docstring += fill(paragraph) + newline + newline
+    return docstring
+
+
+def format_indented_paragraphs(s):
+    return indent(format_paragraphs(s), prefix="    ")
+
+
 def format_parameters(parameters, sig):
     docstring = ""
     # display parameters in order given in function signature
@@ -53,23 +78,51 @@ def format_parameters(parameters, sig):
 
         # handle default value
         if param.default is not Parameter.empty:
-            docstring += f", default={param.default!r}"
+            docstring += ", optional"
+            if param.default is not None:
+                docstring += f", default={param.default!r}"
 
         # add parameter description
         docstring += newline
         param_doc = parameters[param_name]
-        docstring += format_indented_paragraph(param_doc)
+        docstring += format_indented_paragraphs(param_doc).strip(newline)
+        docstring += newline
 
     return docstring
 
 
 def format_type(t):
     # This is probably a bit hacky, could be improved.
-    s = repr(t)
-    if s.startswith("<class"):
-        s = t.__name__
-    s = s.replace("typing.", "")
-    return s
+
+    if t == type(None):  # noqa
+        return "None"
+
+    elif t == ArrayLike:
+        return "array_like"
+
+    elif t == Optional[ArrayLike]:
+        return "array_like or None"
+
+    elif t == DTypeLike:
+        return "data-type"
+
+    elif t == Optional[DTypeLike]:
+        return "data-type or None"
+
+    # special handling for union types
+    elif typing_get_origin(t) == Union and typing_get_args(t):
+        return " or ".join([format_type(x) for x in typing_get_args(t)])
+
+    elif typing_get_origin(t) == Optional and typing_get_args(t):
+        x = typing_get_args(t)[0]
+        return format_type(x) + " or None"
+
+    else:
+        s = repr(t)
+        if s.startswith("<class"):
+            s = t.__name__
+        s = s.replace("typing.", "")
+        return s
 
 
 def format_returns(returns, sig):
@@ -162,9 +215,36 @@ def get_yield_annotation(sig):
         return yield_annotation
 
 
+def get_send_annotation(sig):
+    return_annotation = sig.return_annotation
+
+    if return_annotation is Parameter.empty:
+        # no return annotation
+        return Parameter.empty
+
+    return_type_origin = typing_get_origin(return_annotation)
+
+    # check return type is compatible with yields
+    if return_type_origin is not Generator:
+        raise DocumentationError(
+            f"return type {return_type_origin!r} is not compatible with receives"
+        )
+
+    # extract yield annotation if possible
+    return_type_args = typing_get_args(return_annotation)
+
+    if not return_type_args:
+        # no yield annotation
+        return Parameter.empty
+
+    else:
+        send_annotation = return_type_args[1]
+        return send_annotation
+
+
 def format_yields(yields, sig):
     # yields section is basically the same as the returns section, except we
-    # need to access the yield annotation from within the return annotation
+    # need to access the YieldType annotation from within the return annotation
     if isinstance(yields, str):
         return format_returns_unnamed(yields, get_yield_annotation(sig))
     elif isinstance(yields, Mapping):
@@ -173,6 +253,17 @@ def format_yields(yields, sig):
         raise TypeError("yields must be str or Mapping")
 
 
+def format_receives(receives, sig):
+    # receives section is basically the same as the returns section, except we
+    # need to access the SendType annotation from within the return annotation
+    if isinstance(receives, str):
+        return format_returns_unnamed(receives, get_send_annotation(sig))
+    elif isinstance(receives, Mapping):
+        return format_returns_named(receives, get_send_annotation(sig))
+    else:
+        raise TypeError("receives must be str or Mapping")
+
+
 def format_raises(raises):
     docstring = ""
     for error, description in raises.items():
@@ -181,6 +272,48 @@ def format_raises(raises):
     return docstring
 
 
+def format_maybe_code(obj):
+    return getattr(obj, "__qualname__", str(obj))
+
+
+def format_see_also(see_also):
+    if isinstance(see_also, str):
+        # assume a single function
+        return format_maybe_code(see_also).strip() + newline
+
+    elif isinstance(see_also, Sequence):
+        # assume a sequence of functions
+        docstring = ""
+        for item in see_also:
+            docstring += format_maybe_code(item).strip() + newline
+        return docstring
+
+    elif isinstance(see_also, Mapping):
+        # assume functions with descriptions
+        docstring = ""
+        for name, description in see_also.items():
+            docstring += format_maybe_code(name).strip()
+            if description:
+                docstring += " :"
+                if len(description) < 70:
+                    docstring += " " + description.strip() + newline
+                else:
+                    docstring += newline
+                    docstring += format_indented_paragraph(description)
+            else:
+                docstring += newline
+        return docstring
+
+
+def format_references(references):
+    docstring = ""
+    for ref, desc in references.items():
+        docstring += f".. [{ref}] "
+        desc = format_indented_paragraph(desc).strip()
+        docstring += desc + newline
+    return docstring
+
+
 def doc(
     summary: str,
     deprecation: Optional[Mapping[str, str]] = None,
@@ -188,10 +321,15 @@ def doc(
     parameters: Optional[Mapping[str, str]] = None,
     returns: Optional[Union[str, Mapping[str, str]]] = None,
     yields: Optional[Union[str, Mapping[str, str]]] = None,
+    receives: Optional[Union[str, Mapping[str, str]]] = None,
     other_parameters: Optional[Mapping[str, str]] = None,
     raises: Optional[Mapping[str, str]] = None,
     warns: Optional[Mapping[str, str]] = None,
     warnings: Optional[str] = None,
+    see_also: Optional[Union[str, Sequence[str], Mapping[str, str]]] = None,
+    notes: Optional[str] = None,
+    references: Optional[Mapping[str, str]] = None,
+    examples: Optional[str] = None,
 ):
     """Provide documentation for a function or method, to be formatted as a
     numpy-style docstring (numpydoc).
@@ -206,20 +344,16 @@ def doc(
     extended_summary : str, optional
         A few sentences giving an extended description.
     parameters : Mapping[str, str], optional
-        Description of the function arguments and keywords. Note that types will
-        be obtained from the function's type annotations and do not need to be
-        provided here.
+        Description of the function arguments and keywords.
     returns : str or Mapping[str, str], optional
-        Explanation of the returned values. Note that types will be obtained from
-        the function's return annotation and do not need to be provided here.
+        Explanation of the returned values.
     yields : str or Mapping[str, str], optional
-        Explanation of the yielded values. Note that types will be obtained from
-        the function's return annotation and do not need to be provided here.
+        Explanation of the yielded values.
         This is relevant to generators only.
+    receives : str or Mapping[str, str], optional
+        Explanation of parameters passed to a generator’s `.send()` method.
     other_parameters : Mapping[str, str], optional
-        An optional section used to describe infrequently used parameters. It
-        should only be used if a function has a large number of keyword
-        parameters, to prevent cluttering the Parameters section.
+        An optional section used to describe infrequently used parameters.
     raises : Mapping[str, str], optional
         An optional section detailing which errors get raised and under what
         conditions.
@@ -228,6 +362,15 @@ def doc(
         conditions.
     warnings : str, optional
         An optional section with cautions to the user in free text/reST.
+    see_also : str or Sequence[str] or Mapping[str, str], optional
+        An optional section used to refer to related code.
+    notes : str, optional
+        An optional section that provides additional information about the code,
+        possibly including a discussion of the algorithm.
+    references : Mapping[str, str], optional
+        References cited in the Notes section may be listed here.
+    examples : str, optional
+        An optional section for examples, using the doctest format.
 
     Returns
     -------
@@ -254,6 +397,9 @@ def doc(
     if returns and yields:
         raise DocumentationError("cannot have both returns and yields")
 
+    if receives and not yields:
+        raise DocumentationError("if receives, must also have yields")
+
     def decorator(f):
         docstring = ""
 
@@ -308,6 +454,12 @@ def doc(
             docstring += "------" + newline
             docstring += format_yields(yields, sig)
 
+        # add receives section
+        if receives:
+            docstring += "Receives" + newline
+            docstring += "--------" + newline
+            docstring += format_receives(receives, sig)
+
         # add other parameters section
         if other_parameters:
             docstring += "Other Parameters" + newline
@@ -329,12 +481,39 @@ def doc(
             docstring += format_raises(warns)
             docstring += newline
 
+        # add warnings section
         if warnings:
             docstring += "Warnings" + newline
             docstring += "--------" + newline
             docstring += format_paragraph(warnings)
             docstring += newline
 
+        # add see also section
+        if see_also:
+            docstring += "See Also" + newline
+            docstring += "--------" + newline
+            docstring += format_see_also(see_also)
+            docstring += newline
+
+        # add notes section
+        if notes:
+            docstring += "Notes" + newline
+            docstring += "-----" + newline
+            docstring += format_paragraphs(notes)
+
+        # add references section
+        if references:
+            docstring += "References" + newline
+            docstring += "----------" + newline
+            docstring += format_references(references)
+            docstring += newline
+
+        # add examples section
+        if examples:
+            docstring += "Examples" + newline
+            docstring += "--------" + newline
+            docstring += format_paragraphs(examples)
+
         # final cleanup
         docstring = cleandoc(docstring)
 
